package com.graphaware.neo4j.representation.property;


import org.neo4j.graphdb.PropertyContainer;

import java.util.HashMap;
import java.util.Map;

/**
 * A convenience immutable abstract base-class for Neo4j properties representations that wish to
 * {@link MakesCopyWithProperty} and {@link MakesCopyWithoutProperty}.
 */
public abstract class CopyMakingProperties<T extends Properties> extends StringConvertibleProperties<T> {

    /**
     * Construct a representation of properties from a {@link PropertyContainer}.
     *
     * @param propertyContainer to take (copy) properties from.
     */
    protected CopyMakingProperties(PropertyContainer propertyContainer) {
        super(propertyContainer);
    }

    /**
     * Construct a representation of properties from a {@link Map}.
     *
     * @param properties to take (copy).
     */
    protected CopyMakingProperties(Map<String, String> properties) {
        super(properties);
    }

    /**
     * Construct a representation of properties from a {@link String}.
     *
     * @param string to construct properties from. Must be of the form key1#value1#key2#value2... (assuming the default
     *               {@link com.graphaware.neo4j.utils.Constants#SEPARATOR}.
     */
    protected CopyMakingProperties(String string) {
        super(string);
    }

    /**
     * Make a copy of this instance with a single key-value pair missing.
     *
     * @param key to omit in the copy. If the key does not exist in the representation, an identical copy will be returned.
     * @return copy of this instance except one key-value pair.
     */
    public T without(String key) {
        Map<String, String> newProps = new HashMap<String, String>(getProperties());
        newProps.remove(key);
        return newInstance(newProps);
    }

    /**
     * Make a copy of this instance with an extra key-value pair.
     *
     * @param key   new key
     * @param value new value
     * @return copy of this instance with an extra key-value pair.
     */
    public T with(String key, String value) {
        Map<String, String> newProps = new HashMap<String, String>(getProperties());
        newProps.put(key, value);
        return newInstance(newProps);
    }

    /**
     * Create new instance from a map of properties.
     *
     * @param props props.
     * @return new instance.
     */
    protected abstract T newInstance(Map<String, String> props);
}
