package com.graphaware.neo4j.representation.property;

import java.util.Map;
import java.util.Set;

/**
 * An immutable representation of Neo4j properties, usually coming from a {@link org.neo4j.graphdb.PropertyContainer}.
 * <p/>
 * For the purposes of this representation, keys as well as all values are represented as Strings. Keys must not be null
 * or empty, values can be both, but null values are converted, stored, and treated as empty strings.
 * <p/>
 * Please note that the semantics of {@link #get(String)} and {@link #containsKey(String)} are slightly different than
 * in a regular {@link Map} in the sense that implementations can choose the pretend some key-value pairs that are asked
 * for using these methods exist and generate them. In such cases, all the other methods should still work as expected,
 * i.e. consult the stored properties.
 */
public interface Properties {

    /**
     * Does the instance contain (or can it generate) a property with the given key?
     *
     * @param key to check for.
     * @return true iff it contains the key, or if can generate one.
     */
    boolean containsKey(String key);

    /**
     * Get or generate a property with the given key.
     *
     * @param key key.
     * @return value of the property.
     */
    String get(String key);

    /**
     * Get all the stored keys.
     *
     * @return all keys.
     */
    Set<String> keySet();

    /**
     * Get all stored key-value pairs.
     *
     * @return all entries (key-value pairs).
     */
    Set<Map.Entry<String, String>> entrySet();

    /**
     * @return true iff there are no properties.
     */
    boolean isEmpty();

    /**
     * @return the number of stored properties.
     */
    int size();
}
