package com.graphaware.neo4j.representation.property;


import org.neo4j.graphdb.PropertyContainer;

import java.util.Collections;
import java.util.Map;
import java.util.Set;
import java.util.TreeMap;

import static com.graphaware.neo4j.utils.Constants.SEPARATOR;
import static com.graphaware.neo4j.utils.array.ArrayUtils.isPrimitiveArray;
import static com.graphaware.neo4j.utils.array.ArrayUtils.primitiveArrayToString;

/**
 * A convenience immutable abstract base-class for Neo4j properties representation, capable of converting to and from String.
 * <p/>
 * For the purposes of this representation, keys as well as all values are represented as Strings. Keys must not be null
 * or empty, values can be both, but null values are converted, stored, and treated as empty strings.
 */
public abstract class StringConvertibleProperties<T extends Properties> {

    private final Map<String, String> properties = new TreeMap<String, String>();

    /**
     * Construct a representation of properties from a {@link PropertyContainer}.
     *
     * @param propertyContainer to take (copy) properties from.
     */
    protected StringConvertibleProperties(PropertyContainer propertyContainer) {
        for (String key : propertyContainer.getPropertyKeys()) {
            putProperty(key, valueToString(propertyContainer.getProperty(key)));
        }
    }

    /**
     * Construct a representation of properties from a {@link Map}.
     *
     * @param properties to take (copy).
     */
    protected StringConvertibleProperties(Map<String, String> properties) {
        for (Map.Entry<String, String> entry : properties.entrySet()) {
            putProperty(entry.getKey(), entry.getValue());
        }
    }

    /**
     * Construct a representation of properties from a {@link String}.
     *
     * @param string to construct properties from. Must be of the form key1#value1#key2#value2... (assuming the default
     *               {@link com.graphaware.neo4j.utils.Constants#SEPARATOR}.
     */
    protected StringConvertibleProperties(String string) {
        if (!string.trim().isEmpty()) {
            String[] tokens = string.split(SEPARATOR);
            for (int i = 0; i < tokens.length; i += 2) {
                String key = tokens[i];
                String value = "";
                if (i < tokens.length - 1) {
                    value = tokens[i + 1];
                }

                putProperty(key, value);
            }
        }
    }

    /**
     * Convert a property value to String.
     *
     * @param value to convert.
     * @return property value as String.
     */
    private static String valueToString(Object value) {
        if (isPrimitiveArray(value)) {
            return primitiveArrayToString(value);
        }
        return String.valueOf(value);
    }

    /**
     * Add a new property.
     *
     * @param key   key, must not be null or empty.
     * @param value value, can be null and empty, null will be converted to empty String.
     */
    private void putProperty(String key, String value) {
        if (key == null || "".equals(key.trim())) {
            throw new IllegalArgumentException("Key must not be null or empty!");
        }

        if (value == null) {
            value = "";
        }

        properties.put(key, value);
    }

    /**
     * Does the instance contain a property with the given key?
     *
     * @param key to check for.
     * @return true iff it contains the key.
     */
    public boolean containsKey(String key) {
        return properties.containsKey(key);
    }

    /**
     * Get a property with the given key.
     *
     * @param key key.
     * @return value of the property.
     */
    public String get(String key) {
        return properties.get(key);
    }

    /**
     * Get all the stored keys.
     *
     * @return all keys.
     */
    public Set<String> keySet() {
        return properties.keySet();
    }

    /**
     * Get all stored key-value pairs.
     *
     * @return all entries (key-value pairs).
     */
    public Set<Map.Entry<String,String>> entrySet() {
        return properties.entrySet();
    }

    /**
     * @return true iff there are no properties.
     */
    public boolean isEmpty() {
        return properties.isEmpty();
    }


    /**
     * @return the number of stored properties.
     */
    public int size() {
        return properties.size();
    }

    /**
     * Get an immutable {@link Map} of represented properties.
     *
     * @return read-only {@link Map} of properties.
     */
    public Map<String, String> getProperties() {
        return Collections.unmodifiableMap(properties);
    }

    /**
     * {@inheritDoc}
     */
    @Override
    public String toString() {
        if (properties.isEmpty()) {
            return "";
        }

        StringBuilder stringBuilder = new StringBuilder();
        for (String key : properties.keySet()) {
            stringBuilder
                    .append(key)
                    .append(SEPARATOR)
                    .append(properties.get(key))
                    .append(SEPARATOR);
        }
        return stringBuilder.deleteCharAt(stringBuilder.length() - 1).toString();
    }

    /**
     * {@inheritDoc}
     */
    @Override
    public boolean equals(Object o) {
        if (this == o) return true;
        if (o == null || getClass() != o.getClass()) return false;

        StringConvertibleProperties that = (StringConvertibleProperties) o;

        if (!properties.equals(that.properties)) return false;

        return true;
    }

    /**
     * {@inheritDoc}
     */
    @Override
    public int hashCode() {
        return properties.hashCode();
    }
}
