package com.graphaware.neo4j.representation.relationship;

import com.graphaware.neo4j.representation.property.MakesCopyWithProperty;
import com.graphaware.neo4j.representation.property.MakesCopyWithoutProperty;
import com.graphaware.neo4j.representation.property.Properties;
import org.neo4j.graphdb.*;

import java.util.Map;

/**
 * A convenience immutable abstract base-class for Neo4j relationship representations that wish to
 * {@link MakesCopyWithProperty} and {@link MakesCopyWithoutProperty}.
 *
 * @param <P> type of the {@link com.graphaware.neo4j.representation.property.Properties} stored with this relationship representation.
 */
public abstract class CopyMakingRelationship<P extends Properties & MakesCopyWithProperty<P> & MakesCopyWithoutProperty<P>, R extends Relationship<P>> extends StringConvertibleRelationship<P> {

    /**
     * Construct a relationship representation.
     *
     * @param relationship Neo4j relationship to represent.
     * @param pointOfView  node which is looking at this relationship and thus determines its direction.
     */
    protected CopyMakingRelationship(org.neo4j.graphdb.Relationship relationship, Node pointOfView) {
        super(relationship, pointOfView);
    }

    /**
     * Construct a relationship representation. Please note that using this constructor, the actual properties on the
     * relationship are ignored! The provided properties are used instead.
     *
     * @param relationship Neo4j relationship to represent.
     * @param pointOfView  node which is looking at this relationship and thus determines its direction.
     * @param properties   to use as if they were in the relationship.
     */
    protected CopyMakingRelationship(org.neo4j.graphdb.Relationship relationship, Node pointOfView, P properties) {
        super(relationship, pointOfView, properties);
    }

    /**
     * Construct a relationship representation with no properties.
     *
     * @param type       type.
     * @param direction  direction.
     */
    protected CopyMakingRelationship(RelationshipType type, Direction direction) {
        super(type, direction);
    }

    /**
     * Construct a relationship representation.
     *
     * @param type       type.
     * @param direction  direction.
     * @param properties props.
     */
    protected CopyMakingRelationship(RelationshipType type, Direction direction, P properties) {
        super(type, direction, properties);
    }

    /**
     * Construct a relationship representation.
     *
     * @param type       type.
     * @param direction  direction.
     * @param properties props.
     */
    protected CopyMakingRelationship(RelationshipType type, Direction direction, Map<String, String> properties) {
        super(type, direction, properties);
    }

    /**
     * Construct a relationship representation from a string.
     *
     * @param string string to construct relationship from. Must be of the form type#direction#key1#value1#key2#value2...
     *               (assuming the default {@link com.graphaware.neo4j.utils.Constants#SEPARATOR}.
     */
    protected CopyMakingRelationship(String string) {
        super(string);
    }

    /**
     * Make a copy of this instance with an extra key-value pair.
     *
     * @param key   new key
     * @param value new value
     * @return copy of this instance with an extra key-value pair.
     */
    public R with(String key, String value) {
        return newRelationship(getType(), getDirection(), getProperties().with(key, value));
    }

    /**
     * Make a copy of this instance with a single key-value pair missing.
     *
     * @param key to omit in the copy. If the key does not exist in the representation, an identical copy will be returned.
     * @return copy of this instance except one key-value pair.
     */
    public R without(String key) {
        return newRelationship(getType(), getDirection(), getProperties().without(key));
    }

    /**
     * Create a new instance of this relationship representation.
     *
     * @param type       type.
     * @param direction  direction.
     * @param properties props.
     * @return new instance.
     */
    protected abstract R newRelationship(RelationshipType type, Direction direction, P properties);
}
