package com.graphaware.neo4j.representation.relationship;

import com.graphaware.neo4j.representation.property.Properties;
import org.neo4j.graphdb.*;

import java.util.Map;

import static com.graphaware.neo4j.utils.Constants.GA_REL_PREFIX;
import static com.graphaware.neo4j.utils.Constants.SEPARATOR;
import static com.graphaware.neo4j.utils.relationship.DirectionUtils.resolveDirection;

/**
 * A convenience immutable abstract base-class for Neo4j relationship representation, capable of converting to and from String.
 *
 * @param <P> type of the {@link com.graphaware.neo4j.representation.property.Properties} stored with this relationship representation.
 */
public abstract class StringConvertibleRelationship<P extends Properties> {

    private final RelationshipType type;
    private final Direction direction;
    private final P properties;

    /**
     * Construct a relationship representation.
     *
     * @param relationship Neo4j relationship to represent.
     * @param pointOfView  node which is looking at this relationship and thus determines its direction.
     */
    protected StringConvertibleRelationship(org.neo4j.graphdb.Relationship relationship, Node pointOfView) {
        this.type = relationship.getType();
        this.direction = resolveDirection(relationship, pointOfView);
        this.properties = newProperties(relationship);
    }

    /**
     * Construct a relationship representation. Please note that using this constructor, the actual properties on the
     * relationship are ignored! The provided properties are used instead.
     *
     * @param relationship Neo4j relationship to represent.
     * @param pointOfView  node which is looking at this relationship and thus determines its direction.
     * @param properties   to use as if they were in the relationship.
     */
    protected StringConvertibleRelationship(org.neo4j.graphdb.Relationship relationship, Node pointOfView, P properties) {
        this.type = relationship.getType();
        this.direction = resolveDirection(relationship, pointOfView);
        this.properties = properties;
    }

    /**
     * Construct a relationship representation with no properties.
     *
     * @param type      type.
     * @param direction direction.
     */
    protected StringConvertibleRelationship(RelationshipType type, Direction direction) {
        this.type = type;
        this.direction = direction;
        this.properties = newProperties("");
    }

    /**
     * Construct a relationship representation.
     *
     * @param type       type.
     * @param direction  direction.
     * @param properties props.
     */
    protected StringConvertibleRelationship(RelationshipType type, Direction direction, P properties) {
        this.type = type;
        this.direction = direction;
        this.properties = properties;
    }

    /**
     * Construct a relationship representation.
     *
     * @param type       type.
     * @param direction  direction.
     * @param properties props.
     */
    protected StringConvertibleRelationship(RelationshipType type, Direction direction, Map<String, String> properties) {
        this.type = type;
        this.direction = direction;
        this.properties = newProperties(properties);
    }

    /**
     * Construct a relationship representation from a string.
     *
     * @param string string to construct relationship from. Must be of the form type#direction#key1#value1#key2#value2...
     *               (assuming the default {@link com.graphaware.neo4j.utils.Constants#SEPARATOR}.
     */
    protected StringConvertibleRelationship(String string) {
        if (!string.startsWith(GA_REL_PREFIX)) {
            throw new IllegalArgumentException("This is not a GraphAware Relationship Representation! Must start with " + GA_REL_PREFIX);
        }

        String[] split = string.substring(GA_REL_PREFIX.length()).split(SEPARATOR, 3);
        if (split.length < 2) {
            throw new IllegalArgumentException("Relationship Representation must contain at least type and direction!");
        }

        if (split[0].trim().isEmpty()) {
            throw new IllegalArgumentException("Relationship type must not be empty!");
        }

        this.type = DynamicRelationshipType.withName(split[0]);
        this.direction = Direction.valueOf(split[1]);
        this.properties = newProperties(split.length > 2 ? split[2] : "");
    }

    /**
     * Create properties representation for this relationship from a string.
     *
     * @param string string, as defined in {@link com.graphaware.neo4j.representation.property.StringConvertibleProperties#StringConvertibleProperties(String)}.
     * @return created properties.
     */
    protected abstract P newProperties(String string);


    /**
     * Create properties representation for this relationship from a Neo4j {@link PropertyContainer}.
     *
     * @param propertyContainer to create a representation from.
     * @return created properties.
     */
    protected abstract P newProperties(PropertyContainer propertyContainer);

    /**
     * Create properties representation for this relationship a map of properties..
     *
     * @param properties to create a representation from.
     * @return created properties.
     */
    protected abstract P newProperties(Map<String, String> properties);

    /**
     * Get the relationship type.
     *
     * @return type.
     */
    public RelationshipType getType() {
        return type;
    }

    /**
     * Get the relationship direction.
     *
     * @return direction.
     */
    public Direction getDirection() {
        return direction;
    }

    /**
     * Get the relationship direction.
     *
     * @return direction.
     */
    public P getProperties() {
        return properties;
    }

    /**
     * {@inheritDoc}
     */
    @Override
    public String toString() {
        String result = GA_REL_PREFIX
                + getType().name() + SEPARATOR
                + getDirection().toString() + SEPARATOR
                + getProperties().toString();

        while (result.endsWith(SEPARATOR)) {
            result = result.substring(0, result.length() - 1);
        }

        return result;
    }

    /**
     * {@inheritDoc}
     */
    @Override
    public boolean equals(Object o) {
        if (this == o) return true;
        if (o == null || getClass() != o.getClass()) return false;

        StringConvertibleRelationship that = (StringConvertibleRelationship) o;

        if (direction != that.direction) return false;
        if (!properties.equals(that.properties)) return false;
        if (!type.name().equals(that.type.name())) return false;

        return true;
    }

    /**
     * {@inheritDoc}
     */
    @Override
    public int hashCode() {
        int result = type.name().hashCode();
        result = 31 * result + direction.hashCode();
        result = 31 * result + properties.hashCode();
        return result;
    }
}
