package com.graphaware.neo4j.utils.tx.mutate;

import org.neo4j.graphdb.Relationship;
import org.neo4j.graphdb.event.PropertyEntry;
import org.neo4j.graphdb.event.TransactionData;

import java.util.HashMap;
import java.util.Iterator;
import java.util.Map;

import static com.graphaware.neo4j.utils.iterable.IterableUtils.contains;

/**
 * Static utility class for dealing with various aspects of changing stuff from the database.
 */
public final class ChangeUtils {

    /**
     * Extract changed relationships from {@link org.neo4j.graphdb.event.TransactionData}.
     *
     * @param data about a transaction.
     * @return a map of changed relationships, where the key is the new relationship and the value is original new one.
     *         For information purposes only, these are fake relationships, so don't try to add/remove properties
     *         to/from them thinking that it will be reflected in the database... it won't!
     */
    public static Map<Relationship, Relationship> extractChangedRelationships(TransactionData data) {
        Map<Relationship, Relationship> changedRelationships = new HashMap<Relationship, Relationship>();

        for (PropertyEntry<Relationship> removedProp : data.removedRelationshipProperties()) {
            if (contains(data.deletedRelationships(), removedProp.entity())) {
                continue; //no need to track relationships that were actually deleted, not changed
            }

            Relationship current = new FakeRelationshipWithRealProps(removedProp.entity());
            if (!changedRelationships.containsKey(current)) {
                changedRelationships.put(current, new FakeRelationshipWithRealProps(removedProp.entity()));
            }

            Relationship previous = changedRelationships.get(current);
            previous.setProperty(removedProp.key(), removedProp.previouslyCommitedValue());
        }

        for (PropertyEntry<Relationship> assignedProp : data.assignedRelationshipProperties()) {
            if (contains(data.createdRelationships(), assignedProp.entity())) {
                continue; //no need to track relationships that were actually created, not changed
            }

            Relationship current = new FakeRelationshipWithRealProps(assignedProp.entity());
            if (!changedRelationships.containsKey(current)) {
                changedRelationships.put(current, new FakeRelationshipWithRealProps(assignedProp.entity()));
            }
            Relationship previous = changedRelationships.get(current);
            if (assignedProp.previouslyCommitedValue() == null) {
                previous.removeProperty(assignedProp.key());
            }
            else {
                previous.setProperty(assignedProp.key(), assignedProp.previouslyCommitedValue());
            }
        }

        Iterator<Map.Entry<Relationship,Relationship>> iterator = changedRelationships.entrySet().iterator();
        while (iterator.hasNext()) {
            Map.Entry<Relationship, Relationship> next = iterator.next();
            if (next.getKey().equals(next.getValue())) {
                iterator.remove();
            }
        }

        return changedRelationships;
    }

    private ChangeUtils() {
    }
}
