package com.graphaware.neo4j.utils.tx.mutate;

import org.neo4j.graphdb.Node;
import org.neo4j.graphdb.Relationship;
import org.neo4j.graphdb.event.PropertyEntry;
import org.neo4j.graphdb.event.TransactionData;

import java.util.Collection;
import java.util.HashMap;
import java.util.Map;

/**
 * Static utility class for dealing with various aspects of deleting stuff from the database.
 */
public final class DeleteUtils {

    /**
     * Delete a node, but delete all its relationships first.
     * This method assumes a transaction is in progress.
     *
     * @param toDelete node to delete along with its relationships.
     * @return number of deleted relationships.
     */
    public static int deleteNodeAndRelationships(Node toDelete) {
        int result = 0;
        for (Relationship relationship : toDelete.getRelationships()) {
            relationship.delete();
            result++;
        }
        toDelete.delete();
        return result;
    }

    /**
     * Extract deleted relationships from {@link TransactionData}. These relationships have all their properties accessible,
     * unlike the ones returned by {@link org.neo4j.graphdb.event.TransactionData#deletedRelationships()}.
     *
     * @param data about a transaction.
     * @return a collection of deleted relationships with their properties accessible.
     */
    public static Collection<Relationship> extractDeletedRelationships(TransactionData data) {
        Map<Long, Relationship> deletedRelationships = new HashMap<Long, Relationship>();

        for (Relationship deletedRelationship : data.deletedRelationships()) {
            deletedRelationships.put(deletedRelationship.getId(), new FakeRelationship(deletedRelationship));
        }

        for (PropertyEntry<Relationship> property : data.removedRelationshipProperties()) {
            Relationship fake = deletedRelationships.get(property.entity().getId());
            if (fake != null) {
                fake.setProperty(property.key(), property.previouslyCommitedValue());
            }
        }

        return deletedRelationships.values();
    }

    private DeleteUtils() {
    }
}
